// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
// Basic data structure of an edge
import { EDGE_STATE } from "../core/constants.js";
/** Basic edge data structure */
export class Edge {
    /** Unique uuid of the edge. */
    id;
    /** ID of the source node. */
    _sourceId;
    /** ID of the target node. */
    _targetId;
    /** Whether the edge is directed or not. */
    _directed;
    /** Origin data reference of the edge. */
    _data;
    /** Check the type of the object when picking engine gets it. */
    isEdge = true;
    /** Nodes at either end of this edge. */
    _connectedNodes = {};
    /** Edge state. */
    state = EDGE_STATE.DEFAULT;
    /**
     * The constructor
     * @param options.id - information about the edge
     */
    constructor({ id, sourceId, targetId, data, directed = false }) {
        this.id = id;
        this._sourceId = sourceId;
        this._targetId = targetId;
        this._directed = directed;
        this._data = data;
    }
    /**
     * Return the ID of the edge
     * @return {String|Number} - the ID of the edge.
     */
    getId() {
        return this.id;
    }
    /**
     * Return whether the edge is directed or not.
     * @return {Boolean} true if the edge is directed.
     */
    isDirected() {
        return this._directed;
    }
    /**
     * Get the ID of the source node.
     * @return the ID of the source node.
     */
    getSourceNodeId() {
        return this._sourceId;
    }
    /**
     * Get the ID of the target node.
     * @return the ID of the target node.
     */
    getTargetNodeId() {
        return this._targetId;
    }
    /**
     * Return of the value of the selected property key.
     * @param  key - property key.
     * @return - the value of the property.
     */
    getPropertyValue(key) {
        // try to search the key within this object
        if (this.hasOwnProperty(key)) {
            return this[key];
        }
        // try to search the key in the original data reference
        else if (this._data.hasOwnProperty(key)) {
            return this._data[key];
        }
        // otherwise, not found
        return undefined;
    }
    /**
     * Set the origin data as a reference.
     * @param data - the origin data.
     */
    setData(data) {
        this._data = data;
    }
    /**
     * Update a data property.
     * @param key - the key of the property
     * @param value - the value of the property.
     */
    setDataProperty(key, value) {
        this._data[key] = value;
    }
    /**
     * Set edge state
     * @param state - one of EDGE_STATE
     */
    setState(state) {
        this.state = state;
    }
    /**
     * Get edge state
     * @returns state - one of EDGE_STATE
     */
    getState() {
        return this.state;
    }
    addNode(node) {
        this._connectedNodes[node.getId()] = node;
    }
    removeNode(node) {
        delete this._connectedNodes[node.getId()];
    }
    getConnectedNodes() {
        return Object.values(this._connectedNodes);
    }
}
